#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <xenctrl.h>
#include <xen/io/xenbus.h>
#include <xen/io/protocols.h>
#include <sys/select.h>
#include <stdbool.h>
#include <xen/linux/evtchn.h>
#include <xen/event_channel.h>
#include <sys/mman.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <xs.h>

#include "xenfb.h"
#include "oldxenfb.h"
#include "oldxenkbd.h"

#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))

/* 
 * NB our back-compat hack relies on these structs being the same size
 * or smaller than the same named structs in xenfb.c, and having the
 * same field offsets in both. Do not change this !
 */
struct xenfb_device {
	const char *devicetype;
	char nodename[64];	/* backend xenstore dir */
	char otherend[64];	/* frontend xenstore dir */
	int otherend_id;	/* frontend domid */
	enum xenbus_state state; /* backend state */
	void *page;		/* shared page */
	evtchn_port_t port;
	struct xenfb_private *xenfb;
};

struct xenfb_private {
	struct xenfb pub;
	int evt_xch;		/* event channel driver handle */
	int xc;			/* hypervisor interface handle */
	struct xs_handle *xsh;	/* xs daemon handle */
	struct xenfb_device fb, kbd;
	size_t fb_len;		/* size of framebuffer */
	char protocol[64];	/* frontend protocol */
};

static void oldxenfb_detach_dom(struct xenfb_private *);
static int oldxenfb_fb_event(struct xenfb_private *, union xenfb_in_event *);

static char *oldxenfb_path_in_dom(struct xs_handle *xsh,
				  char *buf, size_t size,
				  unsigned domid, const char *fmt, ...)
{
	va_list ap;
	char *domp = xs_get_domain_path(xsh, domid);
	int n;

        if (domp == NULL)
		return NULL;

	n = snprintf(buf, size, "%s/", domp);
	free(domp);
	if (n >= size)
		return NULL;

	va_start(ap, fmt);
	n += vsnprintf(buf + n, size - n, fmt, ap);
	va_end(ap);
	if (n >= size)
		return NULL;

	return buf;
}

static int oldxenfb_xs_scanf1(struct xs_handle *xsh,
			      const char *dir, const char *node,
			      const char *fmt, void *dest)
{
	char buf[1024];
	char *p;
	int ret;

	if (snprintf(buf, sizeof(buf), "%s/%s", dir, node) >= sizeof(buf)) {
		errno = ENOENT;
		return -1;
        }
	p = xs_read(xsh, XBT_NULL, buf, NULL);
	if (!p) {
		errno = ENOENT;
		return -1;
        }
	ret = sscanf(p, fmt, dest);
	free(p);
	if (ret != 1) {
		errno = EDOM;
		return -1;
        }
	return ret;
}

static int oldxenfb_xs_printf(struct xs_handle *xsh,
			      const char *dir, const char *node, char *fmt, ...)
{
	va_list ap;
	char key[1024];
	char val[1024];
	int n;

	if (snprintf(key, sizeof(key), "%s/%s", dir, node) >= sizeof(key)) {
		errno = ENOENT;
		return -1;
        }

	va_start(ap, fmt);
	n = vsnprintf(val, sizeof(val), fmt, ap);
	va_end(ap);
	if (n >= sizeof(val)) {
		errno = ENOSPC; /* close enough */
		return -1;
	}

	if (!xs_write(xsh, XBT_NULL, key, val, n))
		return -1;
	return 0;
}


static enum xenbus_state oldxenfb_read_state(struct xs_handle *xsh,
					     const char *dir)
{
	int ret, state;

	ret = oldxenfb_xs_scanf1(xsh, dir, "state", "%d", &state);
	if (ret < 0)
		return XenbusStateUnknown;

	if ((unsigned)state > XenbusStateClosed)
		state = XenbusStateUnknown;
	return state;
}

static int oldxenfb_switch_state(struct xenfb_device *dev,
				 enum xenbus_state state)
{
	struct xs_handle *xsh = dev->xenfb->xsh;

	if (oldxenfb_xs_printf(xsh, dev->nodename, "state", "%d", state) < 0)
		return -1;
	dev->state = state;
	return 0;
}

static void oldxenfb_copy_mfns(int mode, int count, unsigned long *dst, void *src)
{
	uint32_t *src32 = src;
	uint64_t *src64 = src;
	int i;

	for (i = 0; i < count; i++)
		dst[i] = (mode == 32) ? src32[i] : src64[i];
}

static int oldxenfb_map_fb(struct xenfb_private *xenfb, int domid)
{
	struct xenfb_page *page = xenfb->fb.page;
	int n_fbmfns;
	int n_fbdirs;
	unsigned long *pgmfns = NULL;
	unsigned long *fbmfns = NULL;
	void *map;
	int mode, ret = -1;

	if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_NATIVE))
	    mode = sizeof(unsigned long) * 8;
	else if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_X86_32))
	    mode = 32;
	else if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_X86_64))
	    mode = 64;
	else
	    return -1;

	n_fbmfns = (xenfb->fb_len + (XC_PAGE_SIZE - 1)) / XC_PAGE_SIZE;
	n_fbdirs = n_fbmfns * mode / 8;
	n_fbdirs = (n_fbdirs + (XC_PAGE_SIZE - 1)) / XC_PAGE_SIZE;

	pgmfns = malloc(sizeof(unsigned long) * n_fbdirs);
	fbmfns = malloc(sizeof(unsigned long) * n_fbmfns);
	if (!pgmfns || !fbmfns)
		goto out;

	/*
	 * Bug alert: xc_map_foreign_batch() can fail partly and
	 * return a non-null value.  This is a design flaw.  When it
	 * happens, we happily continue here, and later crash on
	 * access.
	 */
	oldxenfb_copy_mfns(mode, n_fbdirs, pgmfns, page->pd);
	map = xc_map_foreign_batch(xenfb->xc, domid,
				   PROT_READ, pgmfns, n_fbdirs);
	if (map == NULL)
		goto out;
	oldxenfb_copy_mfns(mode, n_fbmfns, fbmfns, map);
	munmap(map, n_fbdirs * XC_PAGE_SIZE);

	xenfb->pub.pixels = xc_map_foreign_batch(xenfb->xc, domid,
				PROT_READ | PROT_WRITE, fbmfns, n_fbmfns);
	if (xenfb->pub.pixels == NULL)
		goto out;

	ret = 0; /* all is fine */

 out:
	if (pgmfns)
		free(pgmfns);
	if (fbmfns)
		free(fbmfns);
	return ret;
}

static int oldxenfb_bind(struct xenfb_device *dev)
{
	struct xenfb_private *xenfb = dev->xenfb;
	unsigned long mfn;
	evtchn_port_t evtchn;
	char buf[64];
	char *p, **vec;
	unsigned dummy;

	p = oldxenfb_path_in_dom(xenfb->xsh, buf, sizeof(buf),
				 dev->otherend_id, dev->devicetype);
	if (!xs_watch(xenfb->xsh, p, ""))
		return -1;
	for (;;) {
		if (oldxenfb_xs_scanf1(xenfb->xsh, p, "page-ref", "%lu",
				       &mfn) < 0) {
			if (errno == ENOENT || errno == EAGAIN)
				goto wait;
			return -1;
		}
		if (oldxenfb_xs_scanf1(xenfb->xsh, p, "event-channel", "%u",
				       &evtchn) < 0) {
			if (errno == ENOENT || errno == EAGAIN)
				goto wait;
			return -1;
		}
		break;

	wait:
		printf("Waiting...\n");
		vec = xs_read_watch(xenfb->xsh, &dummy);
		if (!vec)
			return -1;
		free(vec);
	}

	dev->port = xc_evtchn_bind_interdomain(xenfb->evt_xch,
					       dev->otherend_id, evtchn);
	if (dev->port == -1)
		return -1;

	dev->page = xc_map_foreign_range(xenfb->xc, dev->otherend_id,
			XC_PAGE_SIZE, PROT_READ | PROT_WRITE, mfn);
	if (dev->page == NULL)
		return -1;

	return 0;
}

static void oldxenfb_unbind(struct xenfb_device *dev)
{
	if (dev->page) {
		munmap(dev->page, XC_PAGE_SIZE);
		dev->page = NULL;
	}
        if (dev->port >= 0) {
		xc_evtchn_unbind(dev->xenfb->evt_xch, dev->port);
		dev->port = -1;
	}
}

static void oldxenfb_dev_fatal(struct xenfb_device *dev, int err,
			       const char *fmt, ...)
{
	struct xs_handle *xsh = dev->xenfb->xsh;
	va_list ap;
	char errdir[80];
	char buf[1024];
	int n;

	fprintf(stderr, "%s ", dev->nodename); /* somewhat crude */
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	if (err)
		fprintf(stderr, " (%s)", strerror(err));
	putc('\n', stderr);

	if (!oldxenfb_path_in_dom(xsh, errdir, sizeof(errdir), 0,
				  "error/%s", dev->nodename))
		goto out;	/* FIXME complain */

	va_start(ap, fmt);
	n = snprintf(buf, sizeof(buf), "%d ", err);
	snprintf(buf + n, sizeof(buf) - n, fmt, ap);
	va_end(ap);

	if (oldxenfb_xs_printf(xsh, buf, "error", "%s", buf) < 0)
		goto out;	/* FIXME complain */

 out:
	oldxenfb_switch_state(dev, XenbusStateClosing);
}

static int xenfb_configure_fb(struct xenfb_private *xenfb,
			      int width, int height, int depth,
			      size_t fb_len, int row_stride)
{
	size_t mfn_sz = sizeof(*((struct xenfb_page *)0)->pd);
	size_t pd_len = sizeof(((struct xenfb_page *)0)->pd) / mfn_sz;
	size_t fb_pages = pd_len * XC_PAGE_SIZE / mfn_sz;
	size_t fb_len_lim = fb_pages * XC_PAGE_SIZE;
	int max_width, max_height;

	if (fb_len > fb_len_lim) {
		fprintf(stderr,
			"FB: frontend fb size %zu limited to %zu\n",
			fb_len, fb_len_lim);
		fb_len = fb_len_lim;
	}
	if (depth != 8 && depth != 16 && depth != 24 && depth != 32) {
		fprintf(stderr,
			"FB: can't handle frontend fb depth %d\n",
			depth);
		return -1;
	}
	if (row_stride < 0 || row_stride > fb_len) {
		fprintf(stderr,
			"FB: invalid frontend stride %d\n", row_stride);
		return -1;
	}
	max_width = row_stride / (depth / 8);
	if (width < 0 || width > max_width) {
		fprintf(stderr,
			"FB: invalid frontend width %d limited to %d\n",
			width, max_width);
		width = max_width;
	}
	max_height = fb_len / row_stride;
	if (height < 0 || height > max_height) {
		fprintf(stderr,
			"FB: invalid frontend height %d limited to %d\n",
			height, max_height);
		height = max_height;
	}
	xenfb->fb_len = fb_len;
	xenfb->pub.row_stride = row_stride;
	xenfb->pub.depth = depth;
	xenfb->pub.width = width;
	xenfb->pub.height = height;
	return 0;
}

int oldxenfb_attach_dom(struct xenfb *xenfb_pub, int domid)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	struct xs_handle *xsh = xenfb->xsh;
	int serrno;
	struct xenfb_page *fb_page;
	union xenfb_in_event event;

	if (oldxenfb_bind(&xenfb->fb) < 0)
	      goto error;
	if (oldxenfb_bind(&xenfb->kbd) < 0)
	      goto error;

	if (oldxenfb_xs_scanf1(xsh, xenfb->fb.otherend, "protocol", "%63s",
			       xenfb->protocol) < 0)
		xenfb->protocol[0] = '\0';

	fb_page = xenfb->fb.page;
	if (xenfb_configure_fb(xenfb,
			       fb_page->width, fb_page->height, fb_page->depth,
			       fb_page->mem_length, fb_page->line_length)
	    < 0) {
		errno = EINVAL;
		goto error;
	}

	if (oldxenfb_map_fb(xenfb, domid) < 0)
		goto error;

	event.type = XENFB_TYPE_SET_EVENTS;
	event.set_events.flags = XENFB_FLAG_UPDATE;
	if (oldxenfb_fb_event(xenfb, &event))
		goto error;

	fprintf(stderr, "Old protocol attach complete\n");
	return 0;

 error:
	fprintf(stderr, "Old protocol attach failed\n");
	serrno = errno;
	oldxenfb_detach_dom(xenfb);
	oldxenfb_dev_fatal(&xenfb->fb, serrno, "on fire");
	oldxenfb_dev_fatal(&xenfb->kbd, serrno, "on fire");
        errno = serrno;
        return -1;
}

static void oldxenfb_detach_dom(struct xenfb_private *xenfb)
{
	oldxenfb_unbind(&xenfb->fb);
	oldxenfb_unbind(&xenfb->kbd);
	if (xenfb->pub.pixels) {
		munmap(xenfb->pub.pixels, xenfb->fb_len);
		xenfb->pub.pixels = NULL;
	}
}

static void oldxenfb_on_fb_event32(struct xenfb_private *xenfb)
{
	uint32_t prod, cons;
	struct xenfb_page32 *page = xenfb->fb.page;
	int x, y, w, h;

	prod = page->out_prod;
	if (prod == page->out_cons)
		return;
	rmb();			/* ensure we see ring contents up to prod */
	for (cons = page->out_cons; cons != prod; cons++) {
		union xenfb_out_event *event = &XENFB_OUT_RING_REF(page, cons);

		switch (event->type) {
		case XENFB_TYPE_UPDATE:
		    x = MAX(event->update.x, 0);
		    y = MAX(event->update.y, 0);
		    w = MIN(event->update.width, xenfb->pub.width - x);
		    h = MIN(event->update.height, xenfb->pub.height - y);
		    if (w < 0 || h < 0) {
			fprintf(stderr, "%s bogus update ignored\n",
				xenfb->fb.nodename);
			break;
		    }
		    if (x != event->update.x || y != event->update.y
			|| w != event->update.width
			|| h != event->update.height) {
			fprintf(stderr, "%s bogus update clipped\n",
				xenfb->fb.nodename);
			break;
		    }
                    if (xenfb->pub.update)
			xenfb->pub.update(&xenfb->pub, x, y, w, h);
                    break;
		}
	}
	mb();			/* ensure we're done with ring contents */
	page->out_cons = cons;
	xc_evtchn_notify(xenfb->evt_xch, xenfb->fb.port);
}

static void oldxenfb_on_fb_event64(struct xenfb_private *xenfb)
{
	uint32_t prod, cons;
	struct xenfb_page64 *page = xenfb->fb.page;
	int x, y, w, h;

	prod = page->out_prod;
	if (prod == page->out_cons)
		return;
	rmb();			/* ensure we see ring contents up to prod */
	for (cons = page->out_cons; cons != prod; cons++) {
		union xenfb_out_event *event = &XENFB_OUT_RING_REF(page, cons);

		switch (event->type) {
		case XENFB_TYPE_UPDATE:
		    x = MAX(event->update.x, 0);
		    y = MAX(event->update.y, 0);
		    w = MIN(event->update.width, xenfb->pub.width - x);
		    h = MIN(event->update.height, xenfb->pub.height - y);
		    if (w < 0 || h < 0) {
			fprintf(stderr, "%s bogus update ignored\n",
				xenfb->fb.nodename);
			break;
		    }
		    if (x != event->update.x || y != event->update.y
			|| w != event->update.width
			|| h != event->update.height) {
			fprintf(stderr, "%s bogus update clipped\n",
				xenfb->fb.nodename);
			break;
		    }
                    if (xenfb->pub.update)
			xenfb->pub.update(&xenfb->pub, x, y, w, h);
                    break;
		}
	}
	mb();			/* ensure we're done with ring contents */
	page->out_cons = cons;
	xc_evtchn_notify(xenfb->evt_xch, xenfb->fb.port);
}

static void oldxenfb_on_fb_event(struct xenfb_private *xenfb)
{
	int mode;
	if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_NATIVE))
	    mode = sizeof(unsigned long) * 8;
	else if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_X86_32))
	    mode = 32;
	else if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_X86_64))
	    mode = 64;
	else
	    return;

	if (mode == 32)
		oldxenfb_on_fb_event32(xenfb);
	else
		oldxenfb_on_fb_event64(xenfb);
}

static void oldxenfb_on_kbd_event(struct xenfb_private *xenfb)
{
	struct xenkbd_info *page = xenfb->kbd.page;

	/* We don't understand any keyboard events, so just ignore them. */
	if (page->out_prod == page->out_cons)
		return;
	page->out_cons = page->out_prod;
	xc_evtchn_notify(xenfb->evt_xch, xenfb->kbd.port);
}

static int oldxenfb_on_state_change(struct xenfb_device *dev)
{
	enum xenbus_state state;

	state = oldxenfb_read_state(dev->xenfb->xsh, dev->otherend);

	switch (state) {
	case XenbusStateUnknown:
		/* There was an error reading the frontend state.  The
		   domain has probably gone away; in any case, there's
		   not much point in us continuing. */
		return -1;
	case XenbusStateInitialising:
	case XenbusStateInitWait:
	case XenbusStateInitialised:
	case XenbusStateConnected:
		break;
	case XenbusStateClosing:
		oldxenfb_unbind(dev);
		oldxenfb_switch_state(dev, state);
		break;
	case XenbusStateClosed:
		oldxenfb_switch_state(dev, state);
	}
	return 0;
}


int oldxenfb_dispatch_channel(struct xenfb *xenfb_pub)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	evtchn_port_t port;
	port = xc_evtchn_pending(xenfb->evt_xch);
	if (port == -1)
		return -1;

	if (port == xenfb->fb.port)
		oldxenfb_on_fb_event(xenfb);
	else if (port == xenfb->kbd.port)
		oldxenfb_on_kbd_event(xenfb);

	if (xc_evtchn_unmask(xenfb->evt_xch, port) == -1)
		return -1;

	return 0;
}

int oldxenfb_dispatch_store(struct xenfb *xenfb_pub)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	unsigned dummy;
	char **vec;
	int r;

	vec = xs_read_watch(xenfb->xsh, &dummy);
	free(vec);
	r = oldxenfb_on_state_change(&xenfb->fb);
	if (r == 0)
		r = oldxenfb_on_state_change(&xenfb->kbd);
	if (r == -1)
		return -2;

	return 0;
}

static int oldxenfb_fb_event32(struct xenfb_private *xenfb,
			       union xenfb_in_event *event)
{
	uint32_t prod;
	struct xenfb_page32 *page = xenfb->fb.page;

	prod = page->in_prod;
	if (prod - page->in_cons == XENFB_IN_RING_LEN) {
		errno = EAGAIN;
		return -1;
	}

	mb();			/* ensure ring space available */
	XENFB_IN_RING_REF(page, prod) = *event;
	wmb();			/* ensure ring contents visible */
	page->in_prod = prod + 1;
	return xc_evtchn_notify(xenfb->evt_xch, xenfb->fb.port);
}

static int oldxenfb_fb_event64(struct xenfb_private *xenfb,
			       union xenfb_in_event *event)
{
	uint32_t prod;
	struct xenfb_page64 *page = xenfb->fb.page;

	prod = page->in_prod;
	if (prod - page->in_cons == XENFB_IN_RING_LEN) {
		errno = EAGAIN;
		return -1;
	}

	mb();			/* ensure ring space available */
	XENFB_IN_RING_REF(page, prod) = *event;
	wmb();			/* ensure ring contents visible */
	page->in_prod = prod + 1;
	return xc_evtchn_notify(xenfb->evt_xch, xenfb->fb.port);
}

static int oldxenfb_fb_event(struct xenfb_private *xenfb,
			     union xenfb_in_event *event)
{
	int mode;
	if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_NATIVE))
	    mode = sizeof(unsigned long) * 8;
	else if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_X86_32))
	    mode = 32;
	else if (0 == strcmp(xenfb->protocol, XEN_IO_PROTO_ABI_X86_64))
	    mode = 64;
	else
	    return -1;

	if (mode == 32)
		return oldxenfb_fb_event32(xenfb, event);
	else
		return oldxenfb_fb_event64(xenfb, event);
}

static int oldxenfb_kbd_event(struct xenfb_private *xenfb,
			      union xenkbd_in_event *event)
{
	uint32_t prod;
	struct xenkbd_info *page = xenfb->kbd.page;

	prod = page->in_prod;
	if (prod - page->in_cons == XENKBD_IN_RING_LEN) {
		errno = EAGAIN;
		return -1;
	}

	mb();			/* ensure ring space available */
	XENKBD_IN_RING_REF(page, prod) = *event;
	wmb();			/* ensure ring contents visible */
	page->in_prod = prod + 1;
	return xc_evtchn_notify(xenfb->evt_xch, xenfb->kbd.port);
}

int oldxenfb_send_key(struct xenfb *xenfb_pub, bool down, int keycode)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	union xenkbd_in_event event;

	memset(&event, 0, XENKBD_IN_EVENT_SIZE);
	event.type = XENKBD_TYPE_KEY;
	event.key.pressed = down ? 1 : 0;
	event.key.keycode = keycode;

	return oldxenfb_kbd_event(xenfb, &event);
}

int oldxenfb_send_motion(struct xenfb *xenfb_pub, int rel_x, int rel_y)
{
	struct xenfb_private *xenfb = (struct xenfb_private *)xenfb_pub;
	union xenkbd_in_event event;

	memset(&event, 0, XENKBD_IN_EVENT_SIZE);
	event.type = XENKBD_TYPE_MOTION;
	event.motion.rel_x = rel_x;
	event.motion.rel_y = rel_y;

	return oldxenfb_kbd_event(xenfb, &event);
}

